import functools
import inspect
import typing
import warnings
from functools import wraps

from docstring_parser import Docstring, parse

from schemafunc.exceptions import (
    BareGenericTypeError,
    NoDocstringError,
    NoShortDescriptionError,
    ParameterMissingDescriptionError,
    ParameterNotDocumentedError,
)
from .type_registry import resolve_type

P = typing.ParamSpec("P")
R = typing.TypeVar("R", covariant=True)


def generate_openai_schema(func: typing.Callable) -> dict:
    return function_to_schema(func)


def generate_anthropic_schema(func: typing.Callable) -> dict:
    openai_schema = function_to_schema(func)
    anthropic_schema = {
        "name": openai_schema["function"]["name"],
        "description": openai_schema["function"]["description"],
        "input_schema": openai_schema["function"]["parameters"],
    }
    # Remove the old keys to move data instead of copying
    del openai_schema["function"]["parameters"]
    del openai_schema["function"]["name"]
    del openai_schema["function"]["description"]
    return anthropic_schema


class OpenAISchema:
    def __init__(self, func: typing.Callable):
        self.func = func

    @functools.cached_property
    def schema(self) -> dict:
        return generate_openai_schema(self.func)


class AnthropicSchema:
    def __init__(self, func: typing.Callable):
        self.func = func

    @functools.cached_property
    def schema(self) -> dict:
        return generate_anthropic_schema(self.func)


class FunctionMetadata:
    def __init__(self, func: typing.Callable[P, R], **schema_kwargs: typing.Any):
        self.func = func
        self.schema_kwargs = schema_kwargs
        self.openai = OpenAISchema(func)
        self.anthropic = AnthropicSchema(func)

    @functools.cached_property
    def schema(self) -> dict:
        warnings.warn(
            "schema is deprecated, use openai.schema instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return function_to_schema(self.func, **self.schema_kwargs)

    @functools.cached_property
    def openai_tool_kwargs(self) -> dict:
        return {
            "tools": [self.schema],
            "tool_choice": {
                "type": "function",
                "function": {"name": self.schema.get("function", {}).get("name")},
            },
        }


class HasSchemaFuncAttribute(typing.Protocol[P, R]):
    schemafunc: FunctionMetadata

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R: ...


DecType = typing.Callable[[typing.Callable[P, R]], HasSchemaFuncAttribute[P, R]]


@typing.overload
def add_schemafunc(
    _func: None = None, **schema_kwargs: typing.Any
) -> DecType[P, R]: ...


@typing.overload
def add_schemafunc(
    _func: typing.Callable[P, R], **schema_kwargs: typing.Any
) -> HasSchemaFuncAttribute[P, R]: ...


def add_schemafunc(
    _func: typing.Optional[typing.Callable[P, R]] = None, **schema_kwargs: typing.Any
) -> typing.Union[DecType[P, R], HasSchemaFuncAttribute[P, R]]:
    def decorator(func: typing.Callable[P, R]) -> HasSchemaFuncAttribute[P, R]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            return func(*args, **kwargs)

        setattr(wrapper, "schemafunc", FunctionMetadata(func, **schema_kwargs))

        # force evaluation so that errors are caught early
        _ = wrapper.schemafunc.schema

        return typing.cast(HasSchemaFuncAttribute[P, R], wrapper)

    if _func is not None:
        return decorator(_func)
    else:
        return decorator


def function_to_schema(
    func: typing.Callable,
    /,
    ignore_args: typing.Sequence[str] = (),
    require_all_params_in_doc: bool = True,
    require_descriptions_for_params: bool = True,
    allow_bare_generic_types: bool = True,
    require_short_description: bool = True,
) -> dict:
    """
    Generates a JSON schema from a given function.

    The function's parameters and their types, as well as the descriptions provided
    in the function's docstring, are included in the resulting schema. This provides
    a machine-readable definition of the function's expected inputs and behaviors.

    Parameters:
        func (typing.Callable): The function to be converted into a schema.
        ignore_args (typing.Sequence[str], optional): A sequence of parameter names
            to ignore when generating the schema. Defaults to ().
        require_all_params_in_doc (bool, optional): If True, an exception is raised
            if not all parameters are documented. Defaults to True.
        require_descriptions_for_params (bool, optional): If True, an exception is
            raised if a documented parameter lacks a description. Defaults to True.
        allow_bare_generic_types (bool, optional): If True, bare generic types such
            as List, Dict are allowed. Defaults to True.
        require_short_description (bool, optional): If True, an exception is raised
            if the function docstring lacks a short description. Defaults to True.

    Returns:
        dict: A representation of the function as a JSON schema.

    Raises:
        NoDocstringError: Raised if the function lacks a docstring and
            require_all_params_in_doc is True.
        NoShortDescriptionError: Raised if function docstring lacks a short
            description and require_short_description is True.
        ValueError: Raised if a parameter lacks a type annotation.
        BareGenericTypeError: Raised if a parameter has a bare generic type and
            allow_bare_generic_types is False.
        ParameterNotDocumentedError: Raised if a parameter is not documented in the
            docstring and require_all_params_in_doc is True.
        ParameterMissingDescriptionError: Raised if a parameter is documented but
            lacks a description and require_descriptions_for_params is True.

    Examples:
        >>> def example_function(param1: int):
        ...    '''
        ...    Simple example function.
        ...    Args:
        ...        param1 (int): The first parameter.
        ...    '''
        ...    pass


        >>> function_to_schema(example_function)
        {'type': 'function', 'function': {'name': 'example_function', 'description': 'Simple example function.', 'parameters': {'type': 'object', 'properties': {'param1': {'type': 'integer', 'description': 'The first parameter.'}}, 'required': ['param1']}}}
    """
    signature = inspect.signature(func)
    filtered_parameters = {
        name: param
        for name, param in signature.parameters.items()
        if name not in ignore_args
    }
    docstring = parse(func.__doc__) if func.__doc__ else None

    if (
        require_all_params_in_doc
        or require_descriptions_for_params
        or require_short_description
    ) and docstring is None:
        raise NoDocstringError("The function must have a docstring")

    if require_short_description and not docstring.short_description:
        raise NoShortDescriptionError(
            "The function must have a short description in the docstring"
        )

    if require_all_params_in_doc and signature.parameters:
        # list of missing parameters. Missing if not in docstring
        missing_params = [
            param_name
            for param_name in signature.parameters
            if param_name not in [p.arg_name for p in getattr(docstring, "params", [])]
        ]
        if missing_params:
            raise ParameterNotDocumentedError(
                "The following parameters are not documented in the docstring: "
                + ", ".join(missing_params)
            )

    if require_descriptions_for_params:
        # list of parameters with missing descriptions
        missing_desc_params = [
            param.arg_name for param in docstring.params if not param.description
        ]

        if missing_desc_params:
            raise ParameterMissingDescriptionError(
                f"The following parameters are missing descriptions in the docstring: {', '.join(missing_desc_params)}"
            )

    parameters = process_parameters(
        filtered_parameters,
        docstring,
        allow_bare_generic_types,
    )
    return generate_schema(func, docstring, parameters, ignore_args)


def process_parameters(
    signature: typing.Mapping[str, inspect.Parameter],
    docstring: Docstring,
    allow_bare_generic_types: bool,
) -> typing.Dict[str, typing.Any]:
    """
    Generates JSON schema representation of the parameters of a function.

    This function systematically processes the parameters of a given function,
    guided by their type annotations and any related documentation provided
    within the function's docstring, and constructs a comprehensive JSON schema
    representation.

    Parameters:
        signature (typing.Mapping[str, inspect.Parameter]): A mapping of parameter
            names to their corresponding inspect.Parameter objects.
        docstring (Docstring): A parsed docstring object which includes
            parameter's documentation.
        allow_bare_generic_types (bool): A boolean parameter that indicates
            whether to allow bare generic types (like List, Dict) without a
            specific item type. If True, it will instantiate the item type
            as an empty object. If False, it will raise a BareGenericTypeError
            for any parameter having a bare generic type.

    Returns:
        dict: A dictionary object that represents the JSON schema of the
            parameters of the function. Each entry represents a parameter,
            with its name as the key and the associated JSON schema as the
            value.

    Raises:
        ValueError: If a parameter lacks a type annotation.
        BareGenericTypeError: If allow_bare_generic_types is False and a
            parameter has a bare generic type.

    Example:
        >>> from inspect import signature
        >>> from docstring_parser import parse

        >>> def example_function(param: int):
        ...     '''
        ...     Example function.
        ...
        ...     Args:
        ...         param: The first parameter.
        ...     '''
        ...     pass
        >>> sig = signature(example_function)
        >>> doc = parse(example_function.__doc__)
        >>> process_parameters(sig.parameters, doc, allow_bare_generic_types=False)
        {'param': {'type': 'integer', 'description': 'The first parameter.'}}

    Note:
        Always provide type annotations for parameters in the function's
        signature. The function relies on these annotations to accurately define
        the JSON schema. The function also leverages the related documentation
        within the function's docstring to enhance the precision of the results,
        providing expanded context when required.
    """
    parameters = {}

    for name, param in signature.items():
        if param.annotation == inspect.Parameter.empty:
            raise ValueError(f"Parameter {name} must have a type annotation")

        param_type = param.annotation
        param_docstring = (
            next((p for p in docstring.params if p.arg_name == name), None)
            if docstring
            else None
        )

        try:
            param_schema = resolve_type(param_type)
        except BareGenericTypeError:
            if allow_bare_generic_types:
                param_schema = {"type": "array", "items": {}}  # Unspecified item type
            else:
                raise

        if param_docstring and param_docstring.description:
            param_schema["description"] = param_docstring.description

        if param.default != inspect.Parameter.empty and param.default is not None:
            param_schema["default"] = param.default

        parameters[name] = param_schema

    return parameters


def generate_schema(
    func: typing.Callable,
    docstring: Docstring,
    parameters: dict,
    ignore_args: typing.Sequence[str] = (),
) -> dict:
    schema = {
        "type": "function",
        "function": {
            "name": func.__name__,
            "description": getattr(docstring, "description", "").strip(),
            "parameters": {
                "type": "object",
                "properties": parameters,
                "required": [
                    name
                    for name, param in inspect.signature(func).parameters.items()
                    if param.default == inspect.Parameter.empty
                    and name not in ignore_args
                ],
            },
        },
    }

    return schema



import types
import typing
from typing import Type

from .registry import TypeHandler, register_type_handler, resolve_type


@register_type_handler
class UnionTypeHandler(TypeHandler):
    def is_type(self, param_type: Type) -> bool:
        return typing.get_origin(param_type) in [typing.Union, types.UnionType]

    def resolve(self, param_type: Type) -> typing.Dict[str, typing.Any]:
        union_types = typing.get_args(param_type)
        resolved_types = [resolve_type(t) for t in union_types]

        schema = {"type": []}
        for resolved_type in resolved_types:
            if resolved_type["type"] == "array":
                schema["type"].append("array")
                schema["items"] = resolved_type["items"]
            else:
                schema["type"].append(resolved_type["type"])

        return schema



from typing import Any, Dict, Literal, Type, get_args, get_origin

from schemafunc.exceptions import UnsupportedLiteralTypeError
from schemafunc.type_registry.basic_types import JSON_SCHEMA_TYPES
from schemafunc.type_registry.registry import TypeHandler, register_type_handler


@register_type_handler
class LiteralTypeHandler(TypeHandler):
    def is_type(self, param_type: Type) -> bool:
        return get_origin(param_type) is Literal

    def resolve(self, param_type: Type) -> Dict[str, Any]:
        literal_values = get_args(param_type)
        # supported_types = (int, float, str, bool, type(None))
        supported_types = tuple(
            x for x in JSON_SCHEMA_TYPES.keys() if isinstance(x, type)
        )
        if all(isinstance(v, supported_types) for v in literal_values):
            return {"type": "string", "enum": list(literal_values)}
        else:
            raise UnsupportedLiteralTypeError(
                f"Unsupported Literal values: {literal_values}"
            )



from types import MappingProxyType
from typing import Any, Dict, Type

from .registry import TypeHandler, register_type_handler

JSON_SCHEMA_TYPES = MappingProxyType(
    {
        int: "integer",
        float: "number",
        str: "string",
        bool: "boolean",
        # Include both None and type(None) in the dictionary to handle two scenarios:
        #
        # 1. When a function parameter is annotated with None directly, such as:
        #      def func(param: None):
        #    The None key is used to map the type to the JSON Schema type "null".
        #
        # 2. When a function parameter is part of a union type with None, such as:
        #      def func(param: typing.Union[int, None]):
        #    The type(None) key is used when resolving the union type to map the None type
        #    to the JSON Schema type "null".
        #
        # By including both None and type(None), the code can handle both scenarios
        # correctly and avoid raising a ValueError for unsupported types.
        None: "null",
        type(None): "null",
    }
)


@register_type_handler
class BasicTypeHandler(TypeHandler):
    def is_type(self, param_type: Type) -> bool:
        return param_type in JSON_SCHEMA_TYPES

    def resolve(self, param_type: Type) -> Dict[str, Any]:
        return {"type": JSON_SCHEMA_TYPES[param_type]}



# array_type_handler.py
from collections.abc import Mapping, Sequence, Set
from typing import Any, Dict, Type, get_args, get_origin

from schemafunc.exceptions import BareGenericTypeError
from schemafunc.type_registry.registry import (
    TypeHandler,
    register_type_handler,
    resolve_type,
)


def is_representable_as_js_array(typ: Type) -> bool:
    """
    Determines if a given type can be accurately represented as a JavaScript array.

    This function returns true if the provided type holds an ordered sequence of
    elements that can be accurately represented as a JavaScript array. Otherwise,
    it returns false.

    Args:
        typ (Type): The Python data type to evaluate.

    Returns:
        bool: Indicates whether the input type can be represented as a JavaScript array.

    Examples:
        >>> is_representable_as_js_array(list)
        True

        >>> is_representable_as_js_array(str)
        False

    """
    try:
        # Check if it's an iterable type, excluding specific types
        # that don't map well to JS arrays.
        if issubclass(typ, (Sequence, Set)) and not issubclass(
            typ, (str, bytes, bytearray, Mapping)
        ):
            return True
    except TypeError:
        # If the type is not a class, then nothing to check.
        return False
    return False


@register_type_handler
class ArrayTypeHandler(TypeHandler):
    def is_type(self, param_type: Type) -> bool:
        return is_representable_as_js_array(get_origin(param_type) or param_type)

    def resolve(self, param_type: Type) -> Dict[str, Any]:
        item_type = get_args(param_type)
        if item_type:
            resolved_item_type = resolve_type(item_type[0])
        else:
            raise BareGenericTypeError(
                f"Bare generic type {param_type} is not allowed."
            )
        return {"type": "array", "items": resolved_item_type}



# mapping_type_handler.py

from collections.abc import Mapping as AbcMapping
from typing import Any, Dict, Mapping, Type, _TypedDictMeta, get_args, get_origin

from schemafunc.exceptions import UnsupportedTypeError
from schemafunc.type_registry.registry import (
    TypeHandler,
    register_type_handler,
    resolve_type,
)


@register_type_handler
class MappingTypeHandler(TypeHandler):
    def is_type(self, param_type: Type) -> bool:
        origin = get_origin(param_type)
        is_typed_dict = isinstance(param_type, _TypedDictMeta)
        if is_typed_dict:
            return False
        valid_origin = origin in [dict, Mapping, AbcMapping]

        try:
            return valid_origin or issubclass(param_type, AbcMapping)
        except TypeError:
            return False

    def resolve(self, param_type: Type) -> Dict[str, Any]:
        key_type, value_type = get_args(param_type)
        if key_type != str:
            raise UnsupportedTypeError(f"Mapping keys must be strings, not {key_type}")
        return {
            "type": "object",
            "additionalProperties": resolve_type(value_type),
        }



from typing import Any, Dict, Type, TypedDict, _TypedDictMeta, get_type_hints

from schemafunc.type_registry.registry import (
    TypeHandler,
    register_type_handler,
    resolve_type,
)


@register_type_handler
class TypedDictTypeHandler(TypeHandler):
    def is_type(self, param_type: Type) -> bool:
        return isinstance(param_type, _TypedDictMeta)

    def resolve(self, param_type: Type[TypedDict]) -> Dict[str, Any]:
        properties = {}
        required = []
        for field_name, field_type in get_type_hints(param_type).items():
            properties[field_name] = resolve_type(field_type)
            if field_name in param_type.__required_keys__:
                required.append(field_name)
        return {
            "type": "object",
            "properties": properties,
            "required": required,
            "additionalProperties": False,
        }



from .array import ArrayTypeHandler
from .basic_types import BasicTypeHandler
from .literal import LiteralTypeHandler
from .mapping import MappingTypeHandler
from .registry import resolve_type
from .typed_dict import TypedDictTypeHandler
from .union import UnionTypeHandler

__all__ = [
    BasicTypeHandler,
    UnionTypeHandler,
    ArrayTypeHandler,
    MappingTypeHandler,
    TypedDictTypeHandler,
    LiteralTypeHandler,
    resolve_type,
]



class NoDocstringError(Exception):
    pass


class ParameterNotDocumentedError(Exception):
    pass


class ParameterMissingDescriptionError(Exception):
    pass


class NoShortDescriptionError(Exception):
    pass


class UnsupportedTypeError(Exception):
    pass


class BareGenericTypeError(ValueError):
    pass


class UnsupportedLiteralTypeError(Exception):
    pass



from abc import ABC
from typing import Any, Dict, Type

from schemafunc.exceptions import UnsupportedTypeError


class TypeHandler(ABC):
    def is_type(self, param_type: Type) -> bool:
        ...

    def resolve(self, param_type: Type) -> Dict[str, Any]:
        ...


type_registry: Dict[TypeHandler, TypeHandler] = {}


def register_type_handler(handler_class: Type[TypeHandler]):
    handler_instance = handler_class()
    type_registry[handler_instance] = handler_instance
    return handler_class


def resolve_type(param_type: Type) -> Dict[str, Any]:
    """
    Resolves a Python type into a JSON Schema dictionary.

    This function takes a Python type (such as `int`, `str`, `Union`,
    `List`, `Dict`, `TypedDict`, etc.) and returns a dictionary
    representing the equivalent JSON Schema type.

    Args:
        param_type (Type): The Python type to resolve.

    Returns:
        dict: A dictionary representing the JSON Schema equivalent of the input type.

    Raises:
        UnsupportedTypeError: If the input type is not supported by the function.

    Examples:
        >>> from typing import *
        >>> resolve_type(int)
        {'type': 'integer'}
    """
    # special cases:
    if param_type is Any:
        return {}
    for handler in type_registry.values():
        if handler.is_type(param_type):
            return handler.resolve(param_type)
    raise UnsupportedTypeError(f"Unsupported type: {param_type}")
